import psycopg2
import random

from django.conf import settings
from django.contrib.auth.models import User
from django.core.management.base import BaseCommand
from django.db import transaction
from django.template.defaultfilters import slugify

from docbow_project.docbow.models import MailingList, DocbowProfile, FileType
from docbow_project.pfwb.models import TabellioDocType


# Generate passwords using capitals and numbers, but no 1 or I, O or 0
__pwd_alphabet = 'ABCDEFGHJKLMNPQRSTUVWXYZ23456789'
def create_password(pwd_length=8):
    password = ''.join([random.choice(__pwd_alphabet) 
        for x in range(pwd_length)])
    return password

def get_username(last_name, first_name):
    username = '%s.%s' % (slugify(first_name), slugify(last_name))
    return username

def get_or_create_user(last_name, first_name, email, verbose):
    try:
        if email == 'noreply@pfwb.be':
            raise User.DoesNotExist()
        user = User.objects.get(email=email)
    except User.DoesNotExist:
        try:
            user = User.objects.get(first_name=first_name, last_name=last_name)
        except User.DoesNotExist:
            user = None

    if user is None:
        if verbose:
            print 'creating account for %s %s <%s>' % (first_name, last_name, email)
        username = get_username(last_name, first_name)
        user, created = User.objects.get_or_create(username=username)
        user.last_name = last_name
        user.first_name = first_name
        user.email = email

    if user.email == 'noreply@pfwb.be':
        # we only set the email when the current email was the placeholder
        # address.
        user.email = email

    user.is_active = True
    user.save()

    # create/update matching docbow profile
    profile, created = DocbowProfile.objects.get_or_create(user=user)

    return user


class Command(BaseCommand):

    @transaction.atomic
    def handle(self, *args, **options):
        verbose = (options.get('verbosity') > 1)

        parl_list = MailingList.objects.get(id=settings.PARLEMENTAIRES_MAILING_ID)
        ministres_list = MailingList.objects.get(id=settings.MINISTRES_MAILING_ID)

        comppols = {}
        commissions_infos = {} # dict of (name, [list of pers id])
        commissions = {} # dict of list of User

        # connect to tabellio database
        conn = psycopg2.connect("host=%s port=%s dbname=%s user=%s password=%s" % (
                settings.TABELLIO_DBHOST,
                settings.TABELLIO_DBPORT,
                settings.TABELLIO_DBNAME,
                settings.TABELLIO_DBUSER,
                settings.TABELLIO_DBPASSWORD))
        cur = conn.cursor()

        # get list of persons that are neither deputy or ministre so they can
        # be disabled if they exists in docbow
        cur.execute('''SELECT t_pers.nom, t_pers.prenom
                         FROM t_pers, t_pershistoline
                        WHERE t_pers.prenom is not NULL
                          AND t_pers.id = t_pershistoline.pers
                          AND t_pershistoline.type IN ('P_CMPL', 'M_MINT')
                          AND t_pershistoline.fin IS NOT NULL
                    ''')

        while True:
            t = cur.fetchone()
            if not t:
                break
            last_name, first_name = t

            username = get_username(last_name, first_name)

            try:
                user = User.objects.get(first_name=first_name, last_name=last_name, is_active=True)
            except User.DoesNotExist:
                # a user didn't exist for that person, fine.
                continue

            if not user.is_active:
                # the user was already disabled, ok.
                continue

            if verbose:
                print 'disabling', username

            # disable the user
            user.is_active = False
            user.save()

            # remove the user from lists
            if user in parl_list.members.all():
                parl_list.members.remove(user)
                parl_list.save()
            if user in ministres_list.members.all():
                ministres_list.members.remove(user)
                ministres_list.save()


        # get current deputies
        deputies = {}

        # in a first pass, get all of them
        cur.execute('''SELECT t_pers.id, t_pers.nom, t_pers.prenom, t_comppol.abbr
                         FROM t_pers, t_pershistoline, t_comppol
                        WHERE t_pers.prenom is not NULL
                          AND t_pershistoline.description = t_comppol.id
                          AND t_pers.id = t_pershistoline.pers
                          AND t_pershistoline.type = 'P_CMPL'
                          AND t_pershistoline.fin IS NULL''')

        while True:
            t = cur.fetchone()
            if not t:
                break
            pers_id, last_name, first_name, comppol = t
            deputies[pers_id] = (last_name, first_name, 'noreply@pfwb.be', comppol)

        # in a second pass, overwrite those who have emails
        cur.execute('''SELECT t_pers.id, t_pers.nom, t_pers.prenom, t_adresse.email, t_comppol.abbr
                         FROM t_pers, t_pershistoline, t_adresse, t_comppol
                        WHERE t_pers.prenom is not NULL
                          AND t_pershistoline.description = t_comppol.id
                          AND t_pers.id = t_pershistoline.pers
                          AND t_pershistoline.type = 'P_CMPL'
                          AND t_pershistoline.fin IS NULL
                          AND t_adresse.email IS NOT NULL
                          AND (t_adresse.id = t_pers.addrpriv
                               OR t_adresse.id = t_pers.addrprof1
                               OR t_adresse.id = t_pers.addrprof2)
                          AND t_adresse.courriel = true''')

        while True:
            t = cur.fetchone()
            if not t:
                break
            pers_id, last_name, first_name, email, comppol = t
            deputies[pers_id] = (last_name, first_name, email, comppol)

        # get commissions
        cur.execute('''SELECT t_com.id, t_com.nom
                            FROM t_com
                           WHERE t_com.st = 'S_ACTIVE'
                           ''')
        while True:
            t = cur.fetchone()
            if not t:
                break
            com_id, nom = t
            commissions_infos[com_id] = (nom, [])
            commissions[com_id] = []

        for com_id in commissions.keys():
            cur.execute('''SELECT t_pers.id
                             FROM t_pers JOIN t_pershistoline
                               ON t_pershistoline.pers = t_pers.id
                            WHERE t_pershistoline.description = %(id)s
                              AND t_pershistoline.fin IS NULL''',
                              {'id': com_id})
            while True:
                t = cur.fetchone()
                if not t:
                    break
                commissions_infos[com_id][1].append(t[0])


        deputy_users = []
        for pers_id, deputy in deputies.items():
            last_name, first_name, email, comppol = deputy

            user = get_or_create_user(last_name, first_name, email, verbose=verbose)
            deputy_users.append(user)

            # update mailing lists
            if not user in parl_list.members.all():
                parl_list.members.add(user)
                parl_list.save()

            # keep track of its political group, to be used later
            if not comppol in comppols:
                comppols[comppol] = []
            comppols[comppol].append(user)

            # keep track of commissions, to be used later
            for com_id, (com_name, com_members) in commissions_infos.items():
                if pers_id in com_members:
                    commissions[com_id].append(user)

        # create mailing lists for political groups
        for comppol, members in comppols.items():
            maillist, created = MailingList.objects.get_or_create(
                    name=u'Appartenance politique - %s' % comppol)
            # remove members of the list that should no longer be in there
            for member in maillist.members.all():
                if member not in members:
                    maillist.members.remove(member)
            # add new members
            for member in members:
                if member not in maillist.members.all():
                    maillist.members.add(member)


        # get inactive comppol
        cur.execute("SELECT abbr FROM t_comppol WHERE st = 'S_INACTIVE'")
        inactive_comppols = [u'Appartenance politique - %s' % i[0] for i in cur.fetchall()]
        comppol_names = [u'Appartenance politique - %s' % comppol for comppol in comppols]
        for maillist in (MailingList.objects.filter(name__startswith='Appartenance politique -')
                         .exclude(name__in=comppol_names)):
            maillist.members.clear()
            if verbose:
                print 'clear', maillist
            if maillist.name in inactive_comppols:
                maillist.is_active = False
                maillist.save()
                if verbose:
                    print 'deactivate', maillist

        existing_commissions = list(MailingList.objects.filter(name__startswith='Commission - '))
        # create mailing lists for commissions
        for com_id, members in commissions.items():
            com_name = commissions_infos[com_id][0]
            maillist, created = MailingList.objects.get_or_create(
                    name=u'Commission - %s' % com_name)
            if maillist in existing_commissions:
                existing_commissions.remove(maillist)
            # remove members of the list that should no longer be in there
            for member in maillist.members.all():
                if member not in members:
                    maillist.members.remove(member)
            # add new members
            for member in members:
                if member not in maillist.members.all():
                    maillist.members.add(member)

        # makes sure remaining commissions are emptied
        for maillist in existing_commissions:
            for member in maillist.members.all():
                maillist.members.remove(member)
            maillist.is_active = False
            maillist.save()

        # get current ministres
        ministres = {}

        # like for deputies, we have a first pass to get all of them
        cur.execute('''SELECT t_pers.id, t_pers.nom, t_pers.prenom
                         FROM t_pers, t_pershistoline
                        WHERE t_pers.prenom is not NULL
                          AND t_pers.id = t_pershistoline.pers
                          AND t_pershistoline.type = 'M_MINT'
                          AND t_pershistoline.fin IS NULL''')

        while True:
            t = cur.fetchone()
            if not t:
                break
            pers_id, last_name, first_name = t
            ministres[pers_id] = (last_name, first_name, 'noreply@pfwb.be')

        # and a second pass to get the email addresses
        cur.execute('''SELECT t_pers.id, t_pers.nom, t_pers.prenom, t_adresse.email
                         FROM t_pers, t_pershistoline, t_adresse
                        WHERE t_pers.st = 'S_MINISTRE'
                          AND t_pers.prenom is not NULL
                          AND t_pers.id = t_pershistoline.pers
                          AND t_pershistoline.type = 'M_MINT'
                          AND t_pershistoline.fin IS NULL
                          AND t_adresse.email IS NOT NULL
                          AND (t_adresse.id = t_pers.addrpriv
                               OR t_adresse.id = t_pers.addrprof1
                               OR t_adresse.id = t_pers.addrprof2)
                          AND t_adresse.courriel = true''')

        while True:
            t = cur.fetchone()
            if not t:
                break
            pers_id, last_name, first_name, email = t
            ministres[pers_id] = (last_name, first_name, email)

        ministres_users = []
        for ministre in ministres.values():
            last_name, first_name, email = ministre

            user = get_or_create_user(last_name, first_name, email, verbose=verbose)
            ministres_users.append(user)

            # update mailing lists
            if not user in ministres_list.members.all():
                ministres_list.members.add(user)
                ministres_list.save()

        # remove users which are not deputies anymore
        for user in parl_list.members.all():
            if user not in deputy_users:
                parl_list.members.remove(user)
                if verbose:
                    print 'remove', user, 'from deputies'

        # remove users which are not ministre anymore
        for user in ministres_list.members.all():
            if user not in ministres_users:
                ministres_list.members.remove(user)
                if verbose:
                    print 'remove', user, 'from ministres'

        # get list of document types
        cur.execute('''SELECT id, descr
                         FROM t_typedoc''')

        while True:
            t = cur.fetchone()
            if not t:
                break
            id, descr = t
            if descr[0] == descr[0].lower():
                descr = descr[0].upper() + descr[1:]

            try:
                doctype = TabellioDocType.objects.get(tabellio_doc_type=id)
            except TabellioDocType.DoesNotExist:
                filetype = FileType()
                filetype.name = descr
                filetype.save()

                doctype = TabellioDocType()
                doctype.tabellio_doc_type = id
                doctype.filetype = filetype
                doctype.save()
            else:
                if doctype.filetype.name != descr:
                    doctype.filetype.name = descr
                    doctype.filetype.save()

        cur.close()
